@@ -8,6 +8,7 @@ from jwcrypto.common import JWSEHeaderParameter, JWSEHeaderRegistry
 from jwcrypto.common import base64url_decode, base64url_encode
 from jwcrypto.common import json_decode, json_encode
 from jwcrypto.jwa import JWA
+from jwcrypto.jwk import JWKSet
 
 
 # RFC 7516 - 4.1
@@ -39,7 +40,7 @@ default_allowed_algs = [
     'ECDH-ES', 'ECDH-ES+A128KW', 'ECDH-ES+A192KW', 'ECDH-ES+A256KW',
     'A128GCMKW', 'A192GCMKW', 'A256GCMKW',
     'PBES2-HS256+A128KW', 'PBES2-HS384+A192KW', 'PBES2-HS512+A256KW',
-    # Content Encryption Algoritms
+    # Content Encryption Algorithms
     'A128CBC-HS256', 'A192CBC-HS384', 'A256CBC-HS512',
     'A128GCM', 'A192GCM', 'A256GCM']
 """Default allowed algorithms"""
@@ -70,7 +71,7 @@ InvalidJWEKeyType = common.InvalidJWEKeyType
 InvalidJWEOperation = common.InvalidJWEOperation
 
 
-class JWE(object):
+class JWE:
     """JSON Web Encryption object
 
     This object represent a JWE token.
@@ -91,7 +92,7 @@ class JWE(object):
         :param header_registry: Optional additions to the header registry
         """
         self._allowed_algs = None
-        self.objects = dict()
+        self.objects = {}
         self.plaintext = None
         self.header_registry = JWSEHeaderRegistry(JWEHeaderRegistry)
         if header_registry:
@@ -164,7 +165,7 @@ class JWE(object):
         return h1
 
     def _get_jose_header(self, header=None):
-        jh = dict()
+        jh = {}
         if 'protected' in self.objects:
             ph = json_decode(self.objects['protected'])
             jh = self._merge_headers(jh, ph)
@@ -229,7 +230,7 @@ class JWE(object):
         jh = self._get_jose_header(header)
         alg, enc = self._get_alg_enc_from_headers(jh)
 
-        rec = dict()
+        rec = {}
         if header:
             rec['header'] = header
 
@@ -250,8 +251,8 @@ class JWE(object):
         if 'recipients' in self.objects:
             self.objects['recipients'].append(rec)
         elif 'encrypted_key' in self.objects or 'header' in self.objects:
-            self.objects['recipients'] = list()
-            n = dict()
+            self.objects['recipients'] = []
+            n = {}
             if 'encrypted_key' in self.objects:
                 n['encrypted_key'] = self.objects.pop('encrypted_key')
             if 'header' in self.objects:
@@ -267,10 +268,13 @@ class JWE(object):
         :param compact(boolean): if True generates the compact
          representation, otherwise generates a standard JSON format.
 
-        :raises InvalidJWEOperation: if the object cannot serialized
+        :raises InvalidJWEOperation: if the object cannot be serialized
          with the compact representation and `compact` is True.
         :raises InvalidJWEOperation: if no recipients have been added
          to the object.
+
+        :return: A json formatted string or a compact representation string
+        :rtype: `str`
         """
 
         if 'ciphertext' not in self.objects:
@@ -280,7 +284,7 @@ class JWE(object):
             for invalid in 'aad', 'unprotected':
                 if invalid in self.objects:
                     raise InvalidJWEOperation(
-                        "Can't use compact encoding when the '%s' parameter"
+                        "Can't use compact encoding when the '%s' parameter "
                         "is set" % invalid)
             if 'protected' not in self.objects:
                 raise InvalidJWEOperation(
@@ -300,7 +304,7 @@ class JWE(object):
                 rec = self.objects
             if 'header' in rec:
                 # The AESGCMKW algorithm generates data (iv, tag) we put in the
-                # per-recipient unpotected header by default. Move it to the
+                # per-recipient unprotected header by default. Move it to the
                 # protected header and re-encrypt the payload, as the protected
                 # header is used as additional authenticated data.
                 h = json_decode(rec['header'])
@@ -329,9 +333,9 @@ class JWE(object):
             if 'aad' in obj:
                 enc['aad'] = base64url_encode(obj['aad'])
             if 'recipients' in obj:
-                enc['recipients'] = list()
+                enc['recipients'] = []
                 for rec in obj['recipients']:
-                    e = dict()
+                    e = {}
                     if 'encrypted_key' in rec:
                         e['encrypted_key'] = \
                             base64url_encode(rec['encrypted_key'])
@@ -355,13 +359,21 @@ class JWE(object):
                     raise InvalidJWEData('Unsupported critical header: '
                                          '"%s"' % k)
 
+    def _unwrap_decrypt(self, alg, enc, key, enckey, header,
+                        aad, iv, ciphertext, tag):
+        cek = alg.unwrap(key, enc.wrap_key_size, enckey, header)
+        data = enc.decrypt(cek, aad, iv, ciphertext, tag)
+        self.decryptlog.append('Success')
+        self.cek = cek
+        return data
+
     # FIXME: allow to specify which algorithms to accept as valid
     def _decrypt(self, key, ppe):
 
         jh = self._get_jose_header(ppe.get('header', None))
 
         # TODO: allow caller to specify list of headers it understands
-        self._check_crit(jh.get('crit', dict()))
+        self._check_crit(jh.get('crit', {}))
 
         for hdr in jh:
             if hdr in self.header_registry:
@@ -374,16 +386,38 @@ class JWE(object):
         aad = base64url_encode(self.objects.get('protected', ''))
         if 'aad' in self.objects:
             aad += '.' + base64url_encode(self.objects['aad'])
+        aad = aad.encode('utf-8')
 
-        cek = alg.unwrap(key, enc.wrap_key_size,
-                         ppe.get('encrypted_key', b''), jh)
-        data = enc.decrypt(cek, aad.encode('utf-8'),
-                           self.objects['iv'],
-                           self.objects['ciphertext'],
-                           self.objects['tag'])
+        if isinstance(key, JWKSet):
+            keys = key
+            if 'kid' in self.jose_header:
+                kid_keys = key.get_keys(self.jose_header['kid'])
+                if not kid_keys:
+                    raise ValueError('Key ID {} not in key set'.format(
+                                     self.jose_header['kid']))
+                keys = kid_keys
 
-        self.decryptlog.append('Success')
-        self.cek = cek
+            for k in keys:
+                try:
+                    data = self._unwrap_decrypt(alg, enc, k,
+                                                ppe.get('encrypted_key', b''),
+                                                jh, aad, self.objects['iv'],
+                                                self.objects['ciphertext'],
+                                                self.objects['tag'])
+                    break
+                except Exception as e:  # pylint: disable=broad-except
+                    keyid = k.get('kid', k.thumbprint())
+                    self.decryptlog.append('Key [{}] failed: [{}]'.format(
+                                           keyid, repr(e)))
+
+            if "Success" not in self.decryptlog:
+                raise ValueError('No working key found in key set')
+        else:
+            data = self._unwrap_decrypt(alg, enc, key,
+                                        ppe.get('encrypted_key', b''),
+                                        jh, aad, self.objects['iv'],
+                                        self.objects['ciphertext'],
+                                        self.objects['tag'])
 
         compress = jh.get('zip', None)
         if compress == 'DEF':
@@ -397,8 +431,9 @@ class JWE(object):
         """Decrypt a JWE token.
 
         :param key: The (:class:`jwcrypto.jwk.JWK`) decryption key.
-        :param key: A (:class:`jwcrypto.jwk.JWK`) decryption key or a password
-         string (optional).
+        :param key: A (:class:`jwcrypto.jwk.JWK`) decryption key,
+         or a (:class:`jwcrypto.jwk.JWKSet`) that contains a key indexed
+         by the 'kid' header or (deprecated) a string containing a password.
 
         :raises InvalidJWEOperation: if the key is not a JWK object.
         :raises InvalidJWEData: if the ciphertext can't be decrypted or
@@ -407,7 +442,7 @@ class JWE(object):
 
         if 'ciphertext' not in self.objects:
             raise InvalidJWEOperation("No available ciphertext")
-        self.decryptlog = list()
+        self.decryptlog = []
 
         if 'recipients' in self.objects:
             for rec in self.objects['recipients']:
@@ -431,22 +466,25 @@ class JWE(object):
         NOTE: Destroys any current status and tries to import the raw
         JWE provided.
 
+        If a key is provided a decryption step will be attempted after
+        the object is successfully deserialized.
+
         :param raw_jwe: a 'raw' JWE token (JSON Encoded or Compact
          notation) string.
-        :param key: A (:class:`jwcrypto.jwk.JWK`) decryption key or a password
-         string (optional).
-         If a key is provided a decryption step will be attempted after
-         the object is successfully deserialized.
+        :param key: A (:class:`jwcrypto.jwk.JWK`) decryption key,
+         or a (:class:`jwcrypto.jwk.JWKSet`) that contains a key indexed
+         by the 'kid' header or (deprecated) a string containing a password
+         (optional).
 
         :raises InvalidJWEData: if the raw object is an invalid JWE token.
         :raises InvalidJWEOperation: if the decryption fails.
         """
 
-        self.objects = dict()
+        self.objects = {}
         self.plaintext = None
         self.cek = None
 
-        o = dict()
+        o = {}
         try:
             try:
                 djwe = json_decode(raw_jwe)
@@ -461,9 +499,9 @@ class JWE(object):
                 if 'aad' in djwe:
                     o['aad'] = base64url_decode(djwe['aad'])
                 if 'recipients' in djwe:
-                    o['recipients'] = list()
+                    o['recipients'] = []
                     for rec in djwe['recipients']:
-                        e = dict()
+                        e = {}
                         if 'encrypted_key' in rec:
                             e['encrypted_key'] = \
                                 base64url_decode(rec['encrypted_key'])
@@ -477,10 +515,10 @@ class JWE(object):
                     if 'header' in djwe:
                         o['header'] = json_encode(djwe['header'])
 
-            except ValueError:
+            except ValueError as e:
                 c = raw_jwe.split('.')
                 if len(c) != 5:
-                    raise InvalidJWEData()
+                    raise InvalidJWEData() from e
                 p = base64url_decode(c[0])
                 o['protected'] = p.decode('utf-8')
                 ekey = base64url_decode(c[1])
@@ -493,7 +531,7 @@ class JWE(object):
             self.objects = o
 
         except Exception as e:  # pylint: disable=broad-except
-            raise InvalidJWEData('Invalid format', repr(e))
+            raise InvalidJWEData('Invalid format', repr(e)) from e
 
         if key:
             self.decrypt(key)
@@ -510,3 +548,52 @@ class JWE(object):
         if len(jh) == 0:
             raise InvalidJWEOperation("JOSE Header not available")
         return jh
+
+    @classmethod
+    def from_jose_token(cls, token):
+        """Creates a JWE object from a serialized JWE token.
+
+        :param token: A string with the json or compat representation
+         of the token.
+
+        :raises InvalidJWEData: if the raw object is an invalid JWE token.
+
+        :return: A JWE token
+        :rtype: JWE
+        """
+
+        obj = cls()
+        obj.deserialize(token)
+        return obj
+
+    def __eq__(self, other):
+        if not isinstance(other, JWE):
+            return False
+        try:
+            return self.serialize() == other.serialize()
+        except Exception:  # pylint: disable=broad-except
+            a = {'plaintext': self.plaintext}
+            a.update(self.objects)
+            b = {'plaintext': other.plaintext}
+            b.update(other.objects)
+            return a == b
+
+    def __str__(self):
+        try:
+            return self.serialize()
+        except Exception:  # pylint: disable=broad-except
+            return self.__repr__()
+
+    def __repr__(self):
+        try:
+            return f'JWE.from_json_token("{self.serialize()}")'
+        except Exception:  # pylint: disable=broad-except
+            plaintext = repr(self.plaintext)
+            protected = self.objects.get('protected')
+            unprotected = self.objects.get('unprotected')
+            aad = self.objects.get('aad')
+            algs = self._allowed_algs
+            return f'JWE(plaintext={plaintext}, ' + \
+                   f'protected={protected}, ' + \
+                   f'unprotected={unprotected}, ' + \
+                   f'aad={aad}, algs={algs})'
