@@ -5,7 +5,7 @@ from jwcrypto.common import JWSEHeaderParameter, JWSEHeaderRegistry
 from jwcrypto.common import base64url_decode, base64url_encode
 from jwcrypto.common import json_decode, json_encode
 from jwcrypto.jwa import JWA
-from jwcrypto.jwk import JWK
+from jwcrypto.jwk import JWK, JWKSet
 
 JWSHeaderRegistry = {
     'alg': JWSEHeaderParameter('Algorithm', False, True, None),
@@ -99,8 +99,9 @@ class JWSCore:
 
         :param alg: The algorithm used to produce the signature.
             See RFC 7518
-        :param key: A (:class:`jwcrypto.jwk.JWK`) key of appropriate
-            type for the "alg" provided in the 'protected' json string.
+        :param key: A (:class:`jwcrypto.jwk.JWK`) verification or
+         a (:class:`jwcrypto.jwk.JWKSet`) that contains a key indexed by the
+         'kid' header. A JWKSet is allowed only for verification operations.
         :param header: A JSON string representing the protected header.
         :param payload(bytes): An arbitrary value
         :param algs: An optional list of allowed algorithms
@@ -112,22 +113,19 @@ class JWSCore:
         """
         self.alg = alg
         self.engine = self._jwa(alg, algs)
-        if not isinstance(key, JWK):
-            raise ValueError('key is not a JWK object')
         self.key = key
 
         if header is not None:
             if isinstance(header, dict):
-                self.header = header
                 header = json_encode(header)
-            else:
-                self.header = json_decode(header)
+            # Make sure this is always a deep copy of the dict
+            self.header = json_decode(header)
 
             self.protected = base64url_encode(header.encode('utf-8'))
         else:
             self.header = {}
             self.protected = ''
-        self.payload = payload
+        self.payload = self._payload(payload)
 
     def _jwa(self, name, allowed):
         if allowed is None:
@@ -136,32 +134,37 @@ class JWSCore:
             raise InvalidJWSOperation('Algorithm not allowed')
         return JWA.signing_alg(name)
 
-    def _payload(self):
+    def _payload(self, payload):
         if self.header.get('b64', True):
-            return base64url_encode(self.payload).encode('utf-8')
+            return base64url_encode(payload).encode('utf-8')
         else:
-            if isinstance(self.payload, bytes):
-                return self.payload
+            if isinstance(payload, bytes):
+                return payload
             else:
-                return self.payload.encode('utf-8')
+                return payload.encode('utf-8')
 
     def sign(self):
         """Generates a signature"""
-        payload = self._payload()
-        sigin = b'.'.join([self.protected.encode('utf-8'), payload])
+        if not isinstance(self.key, JWK):
+            raise ValueError('key is not a JWK object')
+        sigin = b'.'.join([self.protected.encode('utf-8'),
+                           self.payload])
         signature = self.engine.sign(self.key, sigin)
         return {'protected': self.protected,
-                'payload': payload,
+                'payload': self.payload,
                 'signature': base64url_encode(signature)}
 
     def verify(self, signature):
         """Verifies a signature
 
         :raises InvalidJWSSignature: if the verification fails.
+
+        :return: Returns True or an Exception
+        :rtype: `bool`
         """
         try:
-            payload = self._payload()
-            sigin = b'.'.join([self.protected.encode('utf-8'), payload])
+            sigin = b'.'.join([self.protected.encode('utf-8'),
+                               self.payload])
             self.engine.verify(self.key, sigin, signature)
         except Exception as e:  # pylint: disable=broad-except
             raise InvalidJWSSignature('Verification failed') from e
@@ -252,7 +255,6 @@ class JWS:
 
         return header
 
-    # TODO: support selecting key with 'kid' and passing in multiple keys
     def _verify(self, alg, key, payload, signature, protected, header=None):
         p = {}
         # verify it is a valid JSON object and decode
@@ -286,8 +288,33 @@ class JWS:
 
         # the following will verify the "alg" is supported and the signature
         # verifies
-        c = JWSCore(a, key, protected, payload, self._allowed_algs)
-        c.verify(signature)
+        if isinstance(key, JWK):
+            c = JWSCore(a, key, protected, payload, self._allowed_algs)
+            c.verify(signature)
+            self.verifylog.append("Success")
+        elif isinstance(key, JWKSet):
+            keys = key
+            if 'kid' in self.jose_header:
+                kid_keys = key.get_keys(self.jose_header['kid'])
+                if not kid_keys:
+                    raise ValueError('Key ID {} not in key set'.format(
+                                     self.jose_header['kid']))
+                keys = kid_keys
+
+            for k in keys:
+                try:
+                    c = JWSCore(a, k, protected, payload, self._allowed_algs)
+                    c.verify(signature)
+                    self.verifylog.append("Success")
+                    break
+                except Exception as e:  # pylint: disable=broad-except
+                    keyid = k.get('kid', k.thumbprint())
+                    self.verifylog.append('Key [{}] failed: [{}]'.format(
+                                          keyid, repr(e)))
+            if "Success" not in self.verifylog:
+                raise ValueError('No working key found in key set')
+        else:
+            raise ValueError("Unrecognized key type")
 
     # Helper to deal with detached payloads in verification
     def _get_obj_payload(self, obj, dp):
@@ -303,7 +330,9 @@ class JWS:
     def verify(self, key, alg=None, detached_payload=None):
         """Verifies a JWS token.
 
-        :param key: The (:class:`jwcrypto.jwk.JWK`) verification key.
+        :param key: A (:class:`jwcrypto.jwk.JWK`) verification or
+         a (:class:`jwcrypto.jwk.JWKSet`) that contains a key indexed by the
+         'kid' header.
         :param alg: The signing algorithm (optional). Usually the algorithm
             is known as it is provided with the JOSE Headers of the token.
         :param detached_payload: A detached payload to verify the signature
@@ -318,8 +347,8 @@ class JWS:
         self.objects['valid'] = False
         obj = self.objects
         if 'signature' in obj:
+            payload = self._get_obj_payload(obj, detached_payload)
             try:
-                payload = self._get_obj_payload(obj, detached_payload)
                 self._verify(alg, key,
                              payload,
                              obj['signature'],
@@ -330,9 +359,9 @@ class JWS:
                 self.verifylog.append('Failed: [%s]' % repr(e))
 
         elif 'signatures' in obj:
+            payload = self._get_obj_payload(obj, detached_payload)
             for o in obj['signatures']:
                 try:
-                    payload = self._get_obj_payload(obj, detached_payload)
                     self._verify(alg, key,
                                  payload,
                                  o['signature'],
@@ -381,11 +410,14 @@ class JWS:
         NOTE: Destroys any current status and tries to import the raw
         JWS provided.
 
+        If a key is provided a verification step will be attempted after
+        the object is successfully deserialized.
+
         :param raw_jws: a 'raw' JWS token (JSON Encoded or Compact
          notation) string.
-        :param key: A (:class:`jwcrypto.jwk.JWK`) verification key (optional).
-         If a key is provided a verification step will be attempted after
-         the object is successfully deserialized.
+        :param key: A (:class:`jwcrypto.jwk.JWK`) verification or
+         a (:class:`jwcrypto.jwk.JWKSet`) that contains a key indexed by the
+         'kid' header (optional).
         :param alg: The signing algorithm (optional). Usually the algorithm
          is known as it is provided with the JOSE Headers of the token.
 
@@ -454,13 +486,13 @@ class JWS:
 
         b64 = True
 
-        p = {}
         if protected:
             if isinstance(protected, dict):
-                p = protected
-                protected = json_encode(p)
-            else:
-                p = json_decode(protected)
+                protected = json_encode(protected)
+            # Make sure p is always a deep copy of the dict
+            p = json_decode(protected)
+        else:
+            p = dict()
 
         # If b64 is present we must enforce criticality
         if 'b64' in list(p.keys()):
@@ -476,10 +508,9 @@ class JWS:
         h = None
         if header:
             if isinstance(header, dict):
-                h = header
                 header = json_encode(header)
-            else:
-                h = json_decode(header)
+            # Make sure h is always a deep copy of the dict
+            h = json_decode(header)
 
         p = self._merge_check_headers(p, h)
 
@@ -535,6 +566,9 @@ class JWS:
          with the compact representation and `compact` is True.
         :raises InvalidJWSSignature: if no signature has been added
          to the object, or no valid signature can be found.
+
+        :return: A json formatted string or a compact representation string
+        :rtype: `str`
         """
         if compact:
             if 'signatures' in self.objects:
@@ -633,3 +667,41 @@ class JWS:
             return jhl
         else:
             raise InvalidJWSOperation("JOSE Header(s) not available")
+
+    @classmethod
+    def from_jose_token(cls, token):
+        """Creates a JWS object from a serialized JWS token.
+
+        :param token: A string with the json or compat representation
+         of the token.
+
+        :raises InvalidJWSObject: if the raw object is an invalid JWS token.
+
+        :return: A JWS token
+        :rtype: JWS
+        """
+
+        obj = cls()
+        obj.deserialize(token)
+        return obj
+
+    def __eq__(self, other):
+        if not isinstance(other, JWS):
+            return False
+        try:
+            return self.serialize() == other.serialize()
+        except Exception:  # pylint: disable=broad-except
+            return self.objects == other.objects
+
+    def __str__(self):
+        try:
+            return self.serialize()
+        except Exception:  # pylint: disable=broad-except
+            return self.__repr__()
+
+    def __repr__(self):
+        try:
+            return f'JWS.from_json_token("{self.serialize()}")'
+        except Exception:  # pylint: disable=broad-except
+            payload = self.objects['payload'].decode('utf-8')
+            return f'JWS(payload={payload})'
