@@ -1,11 +1,13 @@
 # Copyright (C) 2015  JWCrypto Project Contributors - see LICENSE file
 
+import copy
 import time
 import uuid
 
+from deprecated import deprecated
+
 from jwcrypto.common import JWException, json_decode, json_encode
 from jwcrypto.jwe import JWE
-from jwcrypto.jwk import JWK, JWKSet
 from jwcrypto.jws import JWS
 
 
@@ -106,7 +108,7 @@ class JWTInvalidClaimFormat(JWException):
         super(JWTInvalidClaimFormat, self).__init__(msg)
 
 
-# deprecated and not used anymore
+@deprecated
 class JWTMissingKeyID(JWException):
     """JSON Web Token is missing key id.
 
@@ -125,6 +127,7 @@ class JWTMissingKeyID(JWException):
         super(JWTMissingKeyID, self).__init__(msg)
 
 
+@deprecated
 class JWTMissingKey(JWException):
     """JSON Web Token is using a key not in the key set.
 
@@ -232,15 +235,19 @@ class JWT:
 
     @claims.setter
     def claims(self, c):
-        if self._reg_claims and not isinstance(c, dict):
-            # decode c so we can set default claims
+        if not isinstance(c, dict):
+            if not self._reg_claims:
+                # no default_claims, can return immediately
+                self._claims = c
+                return
             c = json_decode(c)
-
-        if isinstance(c, dict):
-            self._add_default_claims(c)
-            self._claims = json_encode(c)
         else:
-            self._claims = c
+            # _add_default_claims modifies its argument
+            # so we must always copy it.
+            c = copy.deepcopy(c)
+
+        self._add_default_claims(c)
+        self._claims = json_encode(c)
 
     @property
     def token(self):
@@ -358,7 +365,7 @@ class JWT:
     def _check_check_claims(self, check_claims):
         self._check_string_claim('iss', check_claims)
         self._check_string_claim('sub', check_claims)
-        self._check_string_claim('aud', check_claims)
+        self._check_array_or_string_claim('aud', check_claims)
         self._check_integer_claim('exp', check_claims)
         self._check_integer_claim('nbf', check_claims)
         self._check_integer_claim('iat', check_claims)
@@ -398,14 +405,23 @@ class JWT:
 
             elif name == 'aud':
                 if value is not None:
-                    if value == claims[name]:
-                        continue
                     if isinstance(claims[name], list):
-                        if value in claims[name]:
-                            continue
-                    raise JWTInvalidClaimValue(
-                        "Invalid '%s' value. Expected '%s' to be in '%s'" % (
-                            name, claims[name], value))
+                        tclaims = claims[name]
+                    else:
+                        tclaims = [claims[name]]
+                    if isinstance(value, list):
+                        cclaims = value
+                    else:
+                        cclaims = [value]
+                    found = False
+                    for v in cclaims:
+                        if v in tclaims:
+                            found = True
+                            break
+                    if not found:
+                        raise JWTInvalidClaimValue(
+                            "Invalid '{}' value. Expected '{}' in '{}'".format(
+                                name, claims[name], value))
 
             elif name == 'exp':
                 if value is not None:
@@ -434,22 +450,6 @@ class JWT:
                         "Invalid '%s' value. Expected '%s' got '%s'" % (
                             name, value, claims[name]))
 
-    def _deserialize_token_with_keys(self, jwt, keys):
-        for k in keys:
-            try:
-                self.token.deserialize(jwt, k)
-                self.deserializelog.append("Success")
-                break
-            except Exception as e:  # pylint: disable=broad-except
-                keyid = k.get('kid')
-                if keyid is None:
-                    keyid = k.thumbprint()
-                self.deserializelog.append('Key [%s] failed: [%s]' % (
-                    keyid, repr(e)))
-                continue
-        if "Success" not in self.deserializelog:
-            raise JWTMissingKey('No working key found in key set')
-
     def norm_typ(self, val):
         lc = val.lower()
         if '/' in lc:
@@ -489,6 +489,38 @@ class JWT:
         t.add_recipient(key)
         self.token = t
 
+    def validate(self, key):
+        """Validate a JWT token that was deserialized w/o providing a key
+
+        :param key: A (:class:`jwcrypto.jwk.JWK`) verification or
+         decryption key, or a (:class:`jwcrypto.jwk.JWKSet`) that
+         contains a key indexed by the 'kid' header.
+        """
+        self.deserializelog = []
+        if self.token is None:
+            raise ValueError("Token empty")
+
+        try:
+            if isinstance(self.token, JWS):
+                self.token.verify(key)
+            elif isinstance(self.token, JWE):
+                self.token.decrypt(key)
+            else:
+                raise ValueError("Token format unrecognized")
+            self.deserializelog.append("Success")
+        except Exception as e:  # pylint: disable=broad-except
+            if isinstance(self.token, JWS):
+                self.deserializelog = self.token.verifylog
+            elif isinstance(self.token, JWE):
+                self.deserializelog = self.token.decryptlog
+            self.deserializelog.append(
+                'Validation failed: [{}]'.format(repr(e)))
+            raise
+
+        self.header = self.token.jose_header
+        self.claims = self.token.payload.decode('utf-8')
+        self._check_provided_claims()
+
     def deserialize(self, jwt, key=None):
         """Deserialize a JWT token.
 
@@ -512,31 +544,12 @@ class JWT:
         if self._algs:
             self.token.allowed_algs = self._algs
 
-        self.deserializelog = []
+        self.deserializelog = None
         # now deserialize and also decrypt/verify (or raise) if we
         # have a key
-        if key is None:
-            self.token.deserialize(jwt, None)
-        elif isinstance(key, JWK):
-            self.token.deserialize(jwt, key)
-            self.deserializelog.append("Success")
-        elif isinstance(key, JWKSet):
-            self.token.deserialize(jwt, None)
-            if 'kid' in self.token.jose_header:
-                kid_keys = key.get_keys(self.token.jose_header['kid'])
-                if not kid_keys:
-                    raise JWTMissingKey('Key ID %s not in key set'
-                                        % self.token.jose_header['kid'])
-                self._deserialize_token_with_keys(jwt, kid_keys)
-            else:
-                self._deserialize_token_with_keys(jwt, key)
-        else:
-            raise ValueError("Unrecognized Key Type")
-
-        if key is not None:
-            self.header = self.token.jose_header
-            self.claims = self.token.payload.decode('utf-8')
-            self._check_provided_claims()
+        self.token.deserialize(jwt, None)
+        if key:
+            self.validate(key)
 
     def serialize(self, compact=True):
         """Serializes the object into a JWS token.
@@ -548,5 +561,51 @@ class JWT:
         :class:`jwcrypto.jwe.JWE` so that these objects can all be used
         interchangeably. However the only valid JWT representation is the
         compact representation.
+
+        :return: A json formatted string or a compact representation string
+        :rtype: `str`
         """
+        if not compact:
+            raise ValueError("Only the compact serialization is allowed")
+
         return self.token.serialize(compact)
+
+    @classmethod
+    def from_jose_token(cls, token):
+        """Creates a JWT object from a serialized JWT token.
+
+        :param token: A string with the json or compat representation
+         of the token.
+
+        :raises InvalidJWEData or InvalidJWSObject: if the raw object is an
+         invalid JWT token.
+
+        :return: A JWT token
+        :rtype: JWT
+        """
+
+        obj = cls()
+        obj.deserialize(token)
+        return obj
+
+    def __eq__(self, other):
+        if not isinstance(other, JWT):
+            return False
+        return self._claims == other._claims and \
+            self._header == other._header and \
+            self.token == other.token
+
+    def __str__(self):
+        try:
+            return self.serialize()
+        except Exception:  # pylint: disable=broad-except
+            return self.__repr__()
+
+    def __repr__(self):
+        jwt = repr(self.token)
+        return f'JWT(header={self._header}, ' + \
+               f'claims={self._claims}, ' + \
+               f'jwt={jwt}, ' + \
+               f'key=None, algs={self._algs}, ' + \
+               f'default_claims={self._reg_claims}, ' + \
+               f'check_claims={self._check_claims})'
