@@ -995,6 +995,18 @@ class TestJWS(unittest.TestCase):
         t.deserialize(o1)
         t.verify(key)
 
+    def test_jws_issue_281(self):
+        header = {"alg": "HS256"}
+        header_copy = copy.deepcopy(header)
+
+        key = jwk.JWK().generate(kty='oct')
+
+        s = jws.JWS(payload='test')
+        s.add_signature(key, protected=header,
+                        header={"kid": key.thumbprint()})
+
+        self.assertEqual(header, header_copy)
+
 
 E_A1_plaintext = \
     [84, 104, 101, 32, 116, 114, 117, 101, 32, 115, 105, 103, 110, 32,
@@ -1488,7 +1500,7 @@ class TestJWT(unittest.TestCase):
         t.make_encrypted_token(key)
         token = t.serialize()
         # try to decrypt without a matching key
-        self.assertRaises(jwt.JWTMissingKey, jwt.JWT, jwt=token, key=keyset,
+        self.assertRaises(Exception, jwt.JWT, jwt=token, key=keyset,
                           algs=jwe_algs_and_rsa1_5,
                           check_claims={'exp': 1300819380})
         # now decrypt with key
@@ -1502,7 +1514,7 @@ class TestJWT(unittest.TestCase):
         t = jwt.JWT(header, A1_claims, algs=jwe_algs_and_rsa1_5)
         t.make_encrypted_token(key)
         token = t.serialize()
-        self.assertRaises(jwt.JWTMissingKey, jwt.JWT, jwt=token, key=keyset)
+        self.assertRaises(Exception, jwt.JWT, jwt=token, key=keyset)
 
         keyset = jwk.JWKSet.from_json(json_encode(PrivateKeys))
         # encrypt a new JWT with no kid
@@ -1511,7 +1523,7 @@ class TestJWT(unittest.TestCase):
         t.make_encrypted_token(key)
         token = t.serialize()
         # try to decrypt without a matching key
-        self.assertRaises(jwt.JWTMissingKey, jwt.JWT, jwt=token, key=keyset,
+        self.assertRaises(Exception, jwt.JWT, jwt=token, key=keyset,
                           algs=jwe_algs_and_rsa1_5,
                           check_claims={'exp': 1300819380})
         # now decrypt with key
@@ -1534,7 +1546,7 @@ class TestJWT(unittest.TestCase):
         token = t.serialize()
 
         # try to decrypt without a matching key
-        with self.assertRaises(jwt.JWTMissingKey):
+        with self.assertRaises(Exception):
             jwt.JWT(jwt=token, key=keyset, algs=jwe_algs_and_rsa1_5,
                     check_claims={'exp': 1300819380})
 
@@ -1657,14 +1669,27 @@ class TestJWT(unittest.TestCase):
         # the oct key before hitting the ES one
         jwt.JWT(jwt=token, key=ks)
 
-    def test_Issue_239(self):
-        claims = {"aud": "www.example.com"}
-        check_claims = {"aud": ["www.example.com", "account"]}
+    def test_Issue_277(self):
+        claims = {"aud": ["www.example.com", "www.test.net"]}
         key = jwk.JWK(generate='oct', size=256)
         token = jwt.JWT(header={"alg": "HS256"}, claims=claims)
         token.make_signed_token(key)
-        self.assertRaises(jwt.JWTInvalidClaimFormat, jwt.JWT, key=key,
-                          jwt=token.serialize(), check_claims=check_claims)
+        sertok = token.serialize()
+        jwt.JWT(key=key, jwt=sertok, check_claims={"aud": "www.example.com"})
+        jwt.JWT(key=key, jwt=sertok, check_claims={"aud": "www.test.net"})
+        jwt.JWT(key=key, jwt=sertok, check_claims={"aud": ["www.example.com"]})
+        jwt.JWT(key=key, jwt=sertok, check_claims={"aud": ["www.test.net"]})
+        jwt.JWT(key=key, jwt=sertok, check_claims={"aud": ["www.example.com",
+                                                           "www.test.net"]})
+        jwt.JWT(key=key, jwt=sertok, check_claims={"aud": ["www.example.com",
+                                                           "nomatch"]})
+        self.assertRaises(jwt.JWTInvalidClaimValue, jwt.JWT, key=key,
+                          jwt=sertok, check_claims={"aud": "nomatch"})
+        self.assertRaises(jwt.JWTInvalidClaimValue, jwt.JWT, key=key,
+                          jwt=sertok, check_claims={"aud": ["nomatch"]})
+        self.assertRaises(jwt.JWTInvalidClaimValue, jwt.JWT, key=key,
+                          jwt=sertok, check_claims={"aud": ["nomatch",
+                                                            "failmatch"]})
 
 
 class ConformanceTests(unittest.TestCase):
@@ -1788,6 +1813,28 @@ class ConformanceTests(unittest.TestCase):
         check.decrypt(key)
         self.assertEqual(check.payload, b'plain')
 
+    def test_pbes2_hs256_aeskw_custom_params(self):
+        enc = jwe.JWE(plaintext='plain',
+                      protected={"alg": "PBES2-HS256+A128KW",
+                                 "enc": "A256CBC-HS512",
+                                 "p2c": 4096,
+                                 "p2s": base64url_encode("A" * 16)})
+        key = jwk.JWK.from_password('password')
+        enc.add_recipient(key)
+        o = enc.serialize()
+        check = jwe.JWE()
+        check.deserialize(o)
+        check.decrypt(key)
+        self.assertEqual(check.payload, b'plain')
+
+        enc = jwe.JWE(plaintext='plain',
+                      protected={"alg": "PBES2-HS256+A128KW",
+                                 "enc": "A256CBC-HS512",
+                                 "p2c": 4096,
+                                 "p2s": base64url_encode("A" * 7)})
+        key = jwk.JWK.from_password('password')
+        self.assertRaises(ValueError, enc.add_recipient, key)
+
 
 class JWATests(unittest.TestCase):
     def test_jwa_create(self):
@@ -1893,3 +1940,146 @@ class TestUnencodedPayload(unittest.TestCase):
         with self.assertRaises(jws.InvalidJWSObject):
             s.add_signature(jwk.JWK(**SymmetricKeys['keys'][1]),
                             protected=rfc7797_u_header)
+
+
+class TestOverloadedOperators(unittest.TestCase):
+
+    def test_jws_equality(self):
+        key = jwk.JWK.generate(kty='oct', size=256)
+        payload = "My Integrity protected message"
+        a = jws.JWS(payload.encode('utf-8'))
+        b = jws.JWS(payload.encode('utf-8'))
+        self.assertEqual(a, b)
+
+        a.add_signature(key, None,
+                        json_encode({"alg": "HS256"}),
+                        json_encode({"kid": key.thumbprint()}))
+        # One is signed, the other is not
+        self.assertNotEqual(a, b)
+
+        b.add_signature(key, None,
+                        json_encode({"alg": "HS256"}),
+                        json_encode({"kid": key.thumbprint()}))
+        # This kind of signature is deterministic so they should be equal
+        self.assertEqual(a, b)
+
+        c = jws.JWS.from_jose_token(a.serialize())
+        self.assertNotEqual(a, c)
+        c.verify(key)
+        self.assertEqual(a, c)
+
+    def test_jws_representations(self):
+        key = jwk.JWK.generate(kty='oct', size=256)
+        payload = "My Integrity protected message"
+        token = jws.JWS(payload.encode('utf-8'))
+        self.assertEqual(str(token),
+                         "JWS(payload=My Integrity protected message)")
+        self.assertEqual(repr(token),
+                         "JWS(payload=My Integrity protected message)")
+        token.add_signature(key, None,
+                            json_encode({"alg": "HS256"}),
+                            json_encode({"kid": key.thumbprint()}))
+        ser = token.serialize()
+        self.assertEqual(str(token), ser)
+        self.assertEqual(repr(token), f'JWS.from_json_token("{ser}")')
+
+    def test_jwe_equality(self):
+        key = jwk.JWK.generate(kty='oct', size=256)
+        payload = "My Encrypted message"
+        a = jwe.JWE(payload.encode('utf-8'),
+                    json_encode({"alg": "A256KW",
+                                 "enc": "A256CBC-HS512"}))
+        b = jwe.JWE(payload.encode('utf-8'),
+                    json_encode({"alg": "A256KW",
+                                 "enc": "A256CBC-HS512"}))
+        self.assertEqual(a, b)
+
+        a.add_recipient(key)
+        # One is encrypted, the other is not
+        self.assertNotEqual(a, b)
+
+        b.add_recipient(key)
+        # Encryption generates a random CEK so tokens will always differ
+        self.assertNotEqual(a, b)
+
+        c = jwe.JWE.from_jose_token(a.serialize())
+        self.assertEqual(a, c)
+
+    def test_jwe_representations(self):
+        key = jwk.JWK.generate(kty='oct', size=256)
+        payload = "My Encrypted message"
+        token = jwe.JWE(payload.encode('utf-8'),
+                        json_encode({"alg": "A256KW",
+                                     "enc": "A256CBC-HS512"}))
+        strrep = "JWE(plaintext=b\'My Encrypted message\', " + \
+                 "protected={\"alg\":\"A256KW\"," + \
+                 "\"enc\":\"A256CBC-HS512\"}, " + \
+                 "unprotected=None, aad=None, algs=None)"
+        self.assertEqual(str(token), strrep)
+        self.assertEqual(repr(token), strrep)
+
+        token.add_recipient(key)
+        ser = token.serialize()
+        self.assertEqual(str(token), ser)
+        self.assertEqual(repr(token), f'JWE.from_json_token("{ser}")')
+
+    def test_jwt_equality(self):
+        key = jwk.JWK.generate(kty='oct', size=256)
+        a = jwt.JWT(header={"alg": "HS256"},
+                    claims={"info": "I'm a signed token"})
+        b = jwt.JWT(header={"alg": "HS256"},
+                    claims={"info": "I'm a signed token"})
+        self.assertEqual(a, b)
+
+        a.make_signed_token(key)
+        # One is signed, the other is not
+        self.assertNotEqual(a, b)
+
+        b.make_signed_token(key)
+        # This kind of signature is deterministic so they should be equal
+        self.assertEqual(a, b)
+
+        c = jwt.JWT.from_jose_token(a.serialize())
+        self.assertNotEqual(a, c)
+        c.validate(key)
+        self.assertEqual(a, c)
+
+        ea = jwt.JWT(header={"alg": "A256KW", "enc": "A256CBC-HS512"},
+                     claims=a.serialize())
+        eb = jwt.JWT(header={"alg": "A256KW", "enc": "A256CBC-HS512"},
+                     claims=b.serialize())
+        self.assertEqual(ea, eb)
+
+        ea.make_encrypted_token(key)
+        # One is encrypted, the other is not
+        self.assertNotEqual(ea, eb)
+
+        eb.make_encrypted_token(key)
+        # Encryption generates a random CEK so tokens will always differ
+        self.assertNotEqual(ea, eb)
+
+        ect = jwt.JWT.from_jose_token(ea.serialize())
+        self.assertNotEqual(ea, ect)
+        ect.validate(key)
+        self.assertEqual(ea, ect)
+
+    def test_jwt_representations(self):
+        key = jwk.JWK.generate(kty='oct', size=256)
+        token = jwt.JWT(header={"alg": "HS256"},
+                        claims={"info": "I'm a signed token"})
+        strrep = 'JWT(header={"alg":"HS256"}, claims={"info":"I\'m a ' + \
+                 'signed token"}, jwt=None, key=None, algs=None, ' + \
+                 'default_claims=None, check_claims=None)'
+        self.assertEqual(str(token), strrep)
+        self.assertEqual(repr(token), strrep)
+        token.make_signed_token(key)
+
+        ser = token.serialize()
+        self.assertEqual(str(token), ser)
+        ser2 = token.token.serialize()
+
+        reprrep = 'JWT(header={"alg":"HS256"}, ' + \
+                  'claims={"info":"I\'m a signed token"}, ' + \
+                  f'jwt=JWS.from_json_token("{ser2}"), key=None, ' + \
+                  'algs=None, default_claims=None, check_claims=None)'
+        self.assertEqual(repr(token), reprrep)
